from __future__ import annotations

import os
import subprocess
import time
from pathlib import Path
from typing import Dict, List, Optional, Tuple

import docker
import nibabel as nib
import numpy as np
from docker.errors import DockerException
from loguru import logger
from rich.console import Console
from rich.progress import Progress
from rich.table import Table
from rich import box

from brats.constants import DUMMY_PARAMETERS, PARAMETERS_DIR, PACKAGE_CITATION
from brats.utils.algorithm_config import AlgorithmData
from brats.utils.exceptions import (
    AlgorithmNotCPUCompatibleException,
    BraTSContainerException,
)
from brats.utils.zenodo import check_additional_files_path, get_dummy_path

try:
    client = docker.from_env()
except DockerException as e:
    logger.error(
        f"Failed to connect to docker daemon. Please make sure docker is installed and running. Error: {e}"
    )
    # not aborting since this happens during read the docs builds. not a great solution tbf


def _show_docker_pull_progress(tasks: Dict, progress: Progress, line: Dict):
    """Show the progress of a docker pull operation.

    Args:
        tasks (Dict): The tasks to update
        progress (Progress): The progress bar to update
        line (Dict): the next line from docker.client.api.pull stream
    """
    status = line.get("status")
    layer_id = line.get("id", "unknown")

    if status in {"Downloading", "Extracting"}:
        task_key = f"[{status} {layer_id}]"

        progress_detail = line.get("progressDetail") or {}
        total = progress_detail.get("total")
        current = progress_detail.get("current")

        # Case 1: we know total size -> normal progress bar
        if total:
            if task_key not in tasks:
                tasks[task_key] = progress.add_task(f"{task_key}", total=total)
            progress.update(tasks[task_key], completed=current or 0)

        # Case 2: no total/current -> animated bar showing indefinite progress
        else:
            if task_key not in tasks:
                tasks[task_key] = progress.add_task(
                    f"{task_key}", total=None  # total=None means indeterminate
                )
            progress.update(tasks[task_key], advance=0.1)


def _ensure_image(image: str):
    """Ensure the docker image is present on the system. If not, pull it.

    Args:
        image (str): The docker image to pull
    """
    if not client.images.list(name=image):
        logger.info(f"Pulling docker image {image}")
        tasks = {}
        with Progress() as progress:
            resp = client.api.pull(image, stream=True, decode=True)
            for line in resp:
                _show_docker_pull_progress(tasks=tasks, progress=progress, line=line)


def _is_cuda_available() -> bool:
    """Check if CUDA is available on the system by trying to run nvidia-smi."""
    try:
        # Attempt to run `nvidia-smi` to check for CUDA.
        # This command should run successfully if NVIDIA drivers are installed and GPUs are present.
        subprocess.run(
            ["nvidia-smi"],
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
            check=True,
        )
        return True
    except:
        return False


def _handle_device_requests(
    algorithm: AlgorithmData, cuda_devices: str, force_cpu: bool
) -> List[docker.types.DeviceRequest]:
    """Handle the device requests for the docker container (request cuda or cpu).

    Args:
        algorithm (AlgorithmData): Algorithm data
        cuda_devices (str): The CUDA devices to use
        force_cpu (bool): Whether to force CPU execution
    """
    cuda_available = _is_cuda_available()
    if not cuda_available or force_cpu:
        if not algorithm.run_args.cpu_compatible:
            cause = (
                "User tried to force CPU execution but"
                if cuda_available
                else "No Cuda installation/ GPU was found and"
            )
            raise AlgorithmNotCPUCompatibleException(
                f"{cause} the chosen algorithm is not CPU-compatible. Aborting..."
            )
        # empty device requests => run on CPU
        logger.info("Forcing CPU execution")
        return []
    # request gpu with chosen devices
    return [
        docker.types.DeviceRequest(device_ids=[cuda_devices], capabilities=[["gpu"]])
    ]


def _get_additional_files_path(algorithm: AlgorithmData) -> Path:
    """Get the path to the additional files for @algorithm.

    Args:
        algorithm (AlgorithmData): The algorithm data

    Returns:
        Path to the additional files
    """
    # ensure additional_files are present and get path
    if algorithm.additional_files is not None:
        return check_additional_files_path(
            record_id=algorithm.additional_files.record_id
        )
    else:
        # if no additional_files are directly specified a dummy additional_files folder will be mounted
        return get_dummy_path()


def _get_volume_mappings_mlcube(
    data_path: Path,
    additional_files_path: Path,
    output_path: Path,
    parameters_path: Path,
) -> Dict:
    """Get the volume mappings for the docker container.

    Args:
        data_path (Path): The path to the input data
        additional_files_path (Path): The path to the additional files
        output_path (Path): The path to save the output
        parameters_path (Path): The path to mount for the parameters file

    Returns:
        Dict: The volume mappings
    """
    # TODO: add support for recommended "ro" mount mode for input data
    # data = mlcube_io0, additional files = mlcube_io1, output = mlcube_io2, parameters = mlcube_io3
    return {
        volume.absolute(): {
            "bind": f"/mlcube_io{i}",
            "mode": "rw",
        }
        for i, volume in enumerate(
            [data_path, additional_files_path, output_path, parameters_path]
        )
    }


def _get_volume_mappings_docker_only(
    data_path: Path,
    output_path: Path,
) -> Dict[str, Dict[str, str]]:
    """Get the volume mappings for the docker container.

    Args:
        data_path (Path): The path to the input data
        output_path (Path): The path to save the output

    Returns:
        Dict: The volume mappings
    """
    return {
        str(data_path.absolute()): {"bind": "/input", "mode": "rw"},
        str(output_path.absolute()): {"bind": "/output", "mode": "rw"},
    }


def _get_parameters_arg(algorithm: AlgorithmData) -> Optional[str]:
    """Get the parameters argument for the docker container.

    Args:
        algorithm (AlgorithmData): The algorithm data

    Returns:
        Optional[str]: The parameters argument for the docker container or None if a parameter file is not required
    """
    if algorithm.run_args.parameters_file:
        # Docker image name is used as the identifier for the param file
        identifier = algorithm.run_args.docker_image.split(":")[0].split("/")[-1]
        file = PARAMETERS_DIR / f"{identifier}.yml"
        # Some algorithms do require a param file to be present but don't actually use it
        # In this case we simply use a dummy file
        param_file = file if file.exists() else DUMMY_PARAMETERS
        return f" --parameters_file=/mlcube_io3/{param_file.name}"
    return None


def _build_command_args(
    algorithm: AlgorithmData,
) -> str:
    """Build the command arguments for the docker container.

    Args:
        algorithm (AlgorithmData): The algorithm data

    Returns:
        command_args: The command arguments
    """
    # Build command that will be run in the docker container
    command_args = f"--data_path=/mlcube_io0 --output_path=/mlcube_io2"
    if algorithm.additional_files is not None:
        for i, param in enumerate(algorithm.additional_files.param_name):
            additional_files_arg = f"--{param}=/mlcube_io1"
            if algorithm.additional_files.param_path:
                additional_files_arg += f"/{algorithm.additional_files.param_path[i]}"
            command_args += f" {additional_files_arg}"

    # Add parameters file arg if required
    params_arg = _get_parameters_arg(algorithm=algorithm)
    if params_arg:
        command_args += params_arg

    return command_args


def _get_container_user(
    algorithm: AlgorithmData,
) -> Optional[str]:
    """Build extra arguments for the docker container.

    Args:
        algorithm (AlgorithmData): The algorithm data

    Returns:
        Optional[str]: The user to run the container as or None if root is required

    """

    if not algorithm.run_args.requires_root:
        # run the container as the current user to ensure written files are always owned by the user
        # also overall better security-wise
        return f"{os.getuid()}:{os.getgid()}"
    return None


def _observe_docker_output(container: docker.models.containers.Container) -> str:
    """Observe the output of a running docker container and display a spinner. On Errors log container output.

    Args:
        container (docker.models.containers.Container): The container to observe
    """
    # capture the output
    container_output = container.attach(
        stdout=True, stderr=True, stream=True, logs=True
    )

    # Display spinner while the container is running
    with Console().status("Running inference..."):
        # Wait for the container to finish
        exit_code = container.wait()
        container_output = "\n\r".join(
            [line.decode("utf-8", errors="replace") for line in container_output]
        )
        # Check if the container exited with an error
        if exit_code["StatusCode"] != 0:
            logger.error(f">> {container_output}")
            raise BraTSContainerException(
                "Container finished with an error:\n"
                f"{'-'*80}\n{container_output}\n {'-'*80}\n"
                "For further debugging please specify a log file to capture the output or enable console logging (brats.utils.logging.enable())"
            )

    return container_output


def _sanity_check_output(
    data_path: Path,
    output_path: Path,
    container_output: str,
    internal_external_name_map: Optional[Dict[str, str]] = None,
) -> None:
    """Sanity check that the number of output files matches the number of input files and the output is not empty.

    Args:
        data_path (Path): The path to the input data
        output_path (Path): The path to the output data
        container_output (str): The output of the docker container
        internal_external_name_map (Optional[Dict[str, str]]): Dictionary mapping internal name (in standardized format) to external subject name provided by user (only used for batch inference)

    Raises:
        BraTSContainerException: If not enough output files exist
    """

    # some algorithms create extra files in the data folder, so we only check for files starting with "BraTS"
    # (should result in only counting actual inputs)
    inputs = [e for e in data_path.iterdir() if e.name.startswith("BraTS")]
    outputs = list(output_path.iterdir())
    if len(outputs) < len(inputs):
        logger.error(f"Docker container output: \n\r{container_output}")
        raise BraTSContainerException(
            f"Not enough output files were created by the algorithm. Expected: {len(inputs)} Got: {len(outputs)}. Please check the logging output of the docker container for more information."
        )

    for i, output in enumerate(outputs, start=1):
        content = nib.load(output).get_fdata()
        if np.count_nonzero(content) == 0:
            name = "<unknown>"
            if internal_external_name_map is not None:
                name_key = [
                    k for k in internal_external_name_map.keys() if k in output.name
                ]
                if name_key:
                    name = internal_external_name_map[name_key[0]]

            logger.warning(
                f"""Output file for subject {name} contains only zeros.
                Potentially the selected algorithm might not work properly with your data unless this behavior is correct for your use case.
                If this seems wrong please try to use one of the other provided algorithms and file an issue on GitHub if the problem persists."""
            )


def _log_algorithm_info(algorithm: AlgorithmData):
    """Log information about the algorithm and citation reminder.

    Args:
        algorithm (AlgorithmData): algorithm data
    """

    # intentional prints! (to always display them even when logging is disabled)
    console = Console()
    console.rule("[bold red]Citation Reminder[/bold red]")
    console.print(
        "Please support our development by citing the relevant manuscripts for the used algorithm:\n"
    )
    table = Table(
        show_header=False,
        show_lines=True,
        show_edge=False,
        box=box.ASCII,
    )
    table.add_column("type", justify="right", style="cyan")
    table.add_column("url", justify="left", style="white")
    table.add_row("BraTS Package", PACKAGE_CITATION)
    table.add_row(
        f"Challenge ({algorithm.meta.challenge} {algorithm.meta.year})",
        algorithm.meta.challenge_manuscript,
    )
    table.add_row(f"Algorithm ({algorithm.meta.authors})", algorithm.meta.paper)
    if algorithm.meta.dataset_manuscript:
        table.add_row(f"Dataset", algorithm.meta.dataset_manuscript)

    console.print(table)
    console.rule()

    logger.opt(colors=True).info(
        f"Running algorithm: <light-green> BraTS {algorithm.meta.year} {algorithm.meta.challenge} [{algorithm.meta.rank} place]</>"
    )
    logger.debug(f"Docker image: {algorithm.run_args.docker_image}")


def run_container(
    algorithm: AlgorithmData,
    data_path: Path,
    output_path: Path,
    cuda_devices: str,
    force_cpu: bool,
    internal_external_name_map: Optional[Dict[str, str]] = None,
):
    """Run a docker container for the provided algorithm.

    Args:
        algorithm (AlgorithmData): The data of the algorithm to run
        data_path (Path | str): The path to the input data
        output_path (Path | str): The path to save the output
        cuda_devices (str): The CUDA devices to use
        force_cpu (bool): Whether to force CPU execution
        internal_external_name_map (Dict[str, str]): Dictionary mapping internal name (in standardized format) to external subject name provided by user (only used for batch inference)
    """
    _log_algorithm_info(algorithm=algorithm)

    # ensure image is present, if not pull it
    _ensure_image(image=algorithm.run_args.docker_image)

    additional_files_path = _get_additional_files_path(algorithm)

    # ensure output folder exists
    output_path.mkdir(parents=True, exist_ok=True)

    # device setup
    device_requests = _handle_device_requests(
        algorithm=algorithm,
        cuda_devices=cuda_devices,
        force_cpu=force_cpu,
    )
    logger.debug(f"GPU Device requests: {device_requests}")

    user = _get_container_user(algorithm=algorithm)
    logger.debug(f"Container user: {user if user else 'root (required by algorithm)'}")

    logger.info(f"{'Starting inference'}")
    start_time = time.time()

    if algorithm.meta.year <= 2024:
        volume_mappings = _get_volume_mappings_mlcube(
            data_path=data_path,
            additional_files_path=additional_files_path,
            output_path=output_path,
            parameters_path=PARAMETERS_DIR,
        )
        logger.debug(f"Volume mappings: {volume_mappings}")

        command_args = _build_command_args(algorithm=algorithm)
        logger.debug(f"Command args: {command_args}")

        # Run the container
        container = client.containers.run(
            image=algorithm.run_args.docker_image,
            volumes=volume_mappings,
            device_requests=device_requests,
            command=f"infer {command_args}",
            network_mode="none",
            detach=True,
            remove=True,
            shm_size=algorithm.run_args.shm_size,
            user=user,
        )
    else:
        volume_mappings = _get_volume_mappings_docker_only(
            data_path=data_path,
            output_path=output_path,
        )
        logger.debug(f"Volume mappings: {volume_mappings}")
        container = client.containers.run(
            image=algorithm.run_args.docker_image,
            volumes=volume_mappings,
            device_requests=device_requests,
            network_mode="none",
            detach=True,
            remove=True,
            shm_size=algorithm.run_args.shm_size,
            user=user,
        )

    container_output = _observe_docker_output(container=container)
    _sanity_check_output(
        data_path=data_path,
        output_path=output_path,
        container_output=container_output,
        internal_external_name_map=internal_external_name_map,
    )

    logger.debug(f"Docker container output: \n\r{container_output}")

    logger.info(f"Finished inference in {time.time() - start_time:.2f} seconds")



import sys
import threading
from typing import Optional, Union

from loguru import logger

# Singleton handler reference
_console_handler_id: Optional[int] = None
_console_handler_lock = threading.Lock()


def _reset_logging_state_for_tests():
    """
    Reset internal logging state. Only intended for use in tests.
    """
    global _console_handler_id
    logger.remove()
    _console_handler_id = None


def disable():
    """
    Disable the logging for the brats package.
    """

    logger.disable("brats")


def enable():
    """
    Enable the logging for the brats package.
    """

    logger.enable("brats")


def add_console_handler(level: Union[str, int] = "WARNING"):
    """
    Add a console handler to the logger for the brats package.

    Follows the singleton pattern, ensuring only one console handler is present at any time.

    Args:
        level (str | int): The logging level for the console handler. Defaults to "WARNING".
    """
    global _console_handler_id

    with _console_handler_lock:
        if _console_handler_id is None:
            _console_handler_id = logger.add(sys.stderr, level=level)
        else:
            try:
                logger.remove(_console_handler_id)
            except ValueError:
                # If the handler was already removed or doesn't exist, we can safely ignore this.
                pass
            _console_handler_id = logger.add(sys.stderr, level=level)


def remove_console_handler():
    """
    Remove the console handler if it was added.
    """
    global _console_handler_id

    with _console_handler_lock:
        if _console_handler_id is not None:
            try:
                logger.remove(_console_handler_id)
            except ValueError:
                # If the handler was already removed or doesn't exist, we can safely ignore this.
                pass
            _console_handler_id = None



from dacite import DaciteError


class AlgorithmNotCPUCompatibleException(Exception):
    """Exception raised when an CPU-incompatible algorithm tries to run on CPU."""

    pass


class BraTSContainerException(Exception):
    """Exception raised when an algorithm container fails"""

    pass


class AlgorithmConfigException(DaciteError):
    """Exception raised when the algorithm config file has issues."""

    pass


class ZenodoException(Exception):
    """Exception raised when Zenodo is unreachable or returns an error."""

    pass



from brats.utils.algorithm_config import AlgorithmData
from pathlib import Path

import subprocess
from typing import Dict, List, Optional, Union
from brats.core.docker import (
    _log_algorithm_info,
    _sanity_check_output,
    _get_additional_files_path,
    _get_volume_mappings_mlcube,
    _get_parameters_arg,
    _handle_device_requests,
    _get_volume_mappings_docker_only,
    _ensure_image as _ensure_docker_image,
)
from brats.constants import PARAMETERS_DIR
from loguru import logger
import time
from spython.main import Client
import docker
import tempfile
import os

try:
    docker_client = docker.from_env()
except docker.errors.DockerException as e:
    logger.debug(
        "Could not connect to the Docker daemon. Docker functionality is disabled, so the Singularity container's working directory may not be set correctly."
    )
    docker_client = None


def _build_command_args(
    algorithm: AlgorithmData,
) -> List[str]:
    """Build the command arguments for the singularity container.

    Args:
        algorithm (AlgorithmData): The algorithm data

    Returns:
        List[str]: The command arguments
    """

    command_args = ["--data_path=/mlcube_io0", "--output_path=/mlcube_io2"]
    if algorithm.additional_files is not None:
        for i, param in enumerate(algorithm.additional_files.param_name):
            additional_files_arg = f"--{param}=/mlcube_io1"
            if algorithm.additional_files.param_path:
                additional_files_arg += f"/{algorithm.additional_files.param_path[i]}"
            command_args.append(additional_files_arg)

    # Add parameters file arg if required
    params_arg = _get_parameters_arg(algorithm=algorithm)
    if params_arg:
        command_args.append(params_arg.strip())

    return command_args


def _ensure_image(image: str) -> str:
    """
    Ensure the Singularity image is present on the system. If not, pull it as a Sandbox.
    This function checks if the specified Singularity image exists locally in the temporary directory.
    If the image is not found, it pulls the image from Docker Hub, creates a Singularity Sandbox at the target location.

    Args:
        image (str): The Docker image to pull and convert into a Singularity Sandbox.

    Returns:
        str: The path to the Singularity image Sandbox.
    """
    persistent_dir = os.path.join(tempfile.gettempdir(), "brats_singularity_images")
    os.makedirs(persistent_dir, exist_ok=True)
    logger.debug(f"Persistent folder: {persistent_dir}")
    temp_folder = Path(persistent_dir)
    image_path = temp_folder.joinpath(image.replace(":", "_"))
    if not image_path.exists():
        image_path.parent.mkdir(parents=True, exist_ok=True)
        logger.debug(
            f"Pulling Singularity image {image} and creating a Sandbox at {image_path}"
        )
        subprocess.run(
            [
                "singularity",
                "build",
                "--sandbox",
                "--fakeroot",
                str(image_path),
                f"docker://{image}",
            ],
            check=True,
        )

    return str(image_path)


def _convert_volume_mappings_to_singularity_format(
    volume_mappings: Dict[Union[Path, str], Dict[str, str]],
) -> List[str]:
    """Convert volume mappings from Docker format to Singularity format.

    Args:
        volume_mappings (Dict[Path | str, Dict[str, str]]): The volume mappings in Docker format

    Returns:
        List[str]: The volume mappings in Singularity format
    """
    singularity_bindings = []
    for host_path, val in volume_mappings.items():
        container_path = val["bind"]
        singularity_bindings.append(f"{str(host_path)}:{container_path}")
    return singularity_bindings


def _get_docker_working_dir(image: str) -> Optional[Path]:
    """
    Retrieve the working directory configured in the Docker image.

    This is required to properly initialize the working directory for the Singularity container,
    ensuring that the container starts in the correct location as defined by the Docker image.

    Args:
        image (str): The Docker image name or ID.

    Returns:
        Path | None: The working directory specified in the Docker image configuration. None if docker client is not available.
    """
    if docker_client is None:
        return None
    try:
        logger.debug(f"Inspecting image {image}")
        image_obj = docker_client.images.get(image)
    except docker.errors.ImageNotFound:
        logger.debug(f"Image {image} not found locally.")
        _ensure_docker_image(image)
        image_obj = docker_client.images.get(image)
    workdir = image_obj.attrs["Config"].get("WorkingDir", None)
    logger.debug(f"Working directory: {workdir}")
    if workdir is None:
        return None
    else:
        return Path(workdir)


def run_container(
    algorithm: AlgorithmData,
    data_path: Path,
    output_path: Path,
    cuda_devices: str,
    force_cpu: bool,
    internal_external_name_map: Optional[Dict[str, str]] = None,
    overlay_size: int = 1024,
):
    """Run a Singularity container for the provided algorithm.

    Args:
        algorithm (AlgorithmData): The data of the algorithm to run
        data_path (Path | str): The path to the input data
        output_path (Path | str): The path to save the output
        cuda_devices (str): The CUDA devices to use
        force_cpu (bool): Whether to force CPU execution
        internal_external_name_map (Dict[str, str]): Dictionary mapping internal name (in standardized format) to external subject name provided by user (only used for batch inference)
        overlay_size (int): The size of the overlay image in MB. Defaults to 1024.
    """
    if overlay_size <= 0:
        raise ValueError("Overlay size must be greater than 0.")
    _log_algorithm_info(algorithm=algorithm)
    # ensure image is present, if not pull it
    image = _ensure_image(image=algorithm.run_args.docker_image)

    additional_files_path = _get_additional_files_path(algorithm)
    logger.debug(f"Additional files path: {additional_files_path}")
    # ensure output folder exists
    output_path.mkdir(parents=True, exist_ok=True)

    command_args = _build_command_args(algorithm=algorithm)
    command_args_str = " ".join(command_args)
    logger.debug(f"Command args: {command_args_str}")
    if algorithm.meta.year <= 2024:
        volume_mappings = _get_volume_mappings_mlcube(
            data_path=data_path,
            additional_files_path=additional_files_path,
            output_path=output_path,
            parameters_path=PARAMETERS_DIR,
        )
        args = ["infer", *command_args]
    else:
        volume_mappings = _get_volume_mappings_docker_only(
            data_path=data_path,
            output_path=output_path,
        )
        args = None

    logger.debug(f"Volume mappings: {volume_mappings}")

    # device setup
    device_requests = _handle_device_requests(
        algorithm=algorithm, cuda_devices=cuda_devices, force_cpu=force_cpu
    )
    logger.debug(f"GPU Device requests: {device_requests}")

    # Run the container
    logger.info("Starting inference")
    start_time = time.time()

    singularity_bindings = _convert_volume_mappings_to_singularity_format(
        volume_mappings
    )

    options = []

    if len(device_requests) > 0 and not force_cpu:
        logger.info(f"Using CUDA devices: {cuda_devices}")
        options.append("--nv")  # Singularity uses --nv to enable GPU support

    # TODO: The --fakeroot option may be required for certain algorithms that need root privileges inside the Singularity container.
    docker_working_dir = _get_docker_working_dir(algorithm.run_args.docker_image)
    if docker_working_dir is not None:
        options.append("--cwd")
        options.append(str(docker_working_dir))
    else:
        logger.warning(
            "Docker working directory not found. Using default working directory."
        )
    overlay_path = Path(image).parent / (Path(image).name + "_overlay.img")
    options.append("--overlay")
    options.append(str(overlay_path))

    overlay_created = False
    if not overlay_path.exists():
        subprocess.run(
            [
                "singularity",
                "overlay",
                "create",
                "--size",
                str(overlay_size),
                str(overlay_path),
            ],
            check=True,
        )
        overlay_created = True
    try:
        executor = Client.run(
            image,
            options=options,
            args=args,
            stream=True,
            bind=singularity_bindings,
        )
        container_output = []
        for line in executor:
            container_output.append(line)

        _sanity_check_output(
            data_path=data_path,
            output_path=output_path,
            container_output="\n".join(container_output),
            internal_external_name_map=internal_external_name_map,
        )
    finally:
        try:
            if overlay_created:
                overlay_path.unlink()
        except FileNotFoundError:
            pass
    logger.info(f"Finished inference in {time.time() - start_time:.2f} seconds")



from __future__ import annotations

import shutil
import sys
import zipfile
from io import BytesIO
from pathlib import Path
from typing import Dict, List, Tuple

import requests
from loguru import logger
from rich.progress import Progress, SpinnerColumn, TextColumn

from brats.constants import ADDITIONAL_FILES_FOLDER, ZENODO_RECORD_BASE_URL
from brats.utils.exceptions import ZenodoException


def get_dummy_path() -> Path:
    dummy = ADDITIONAL_FILES_FOLDER / "dummy"
    dummy.mkdir(exist_ok=True, parents=True)
    return dummy


def check_additional_files_path(record_id: str) -> Path:
    """Check if latest additional files are present and download them otherwise.

    Args:
        record_id (str): Zenodo record ID.

    Raises:
        ZenodoException: If the additional files are not found locally and Zenodo is unreachable.

    Returns:
        Path: Path to the additional files folder.
    """

    zenodo_response = _get_zenodo_metadata_and_archive_url(record_id=record_id)

    record_additional_files_pattern = f"{record_id}_v*.*.*"
    matching_folders = list(
        ADDITIONAL_FILES_FOLDER.glob(record_additional_files_pattern)
    )
    # Get the latest downloaded additional_files
    latest_downloaded_additional_files = _get_latest_version_folder_name(
        matching_folders
    )

    if not latest_downloaded_additional_files:
        if not zenodo_response:
            msg = "Additional files not found locally and Zenodo could not be reached. Exiting..."
            logger.error(msg)
            raise ZenodoException(msg)
        logger.info(f"Additional files not found locally")

        zenodo_metadata, archive_url = zenodo_response
        return _download_additional_files(
            zenodo_metadata=zenodo_metadata,
            record_id=record_id,
            archive_url=archive_url,
        )

    logger.info(
        f"Found downloaded local additional_files: {latest_downloaded_additional_files}"
    )

    if not zenodo_response:
        logger.warning(
            "Zenodo server could not be reached. Using the latest downloaded additional files."
        )
        return ADDITIONAL_FILES_FOLDER / latest_downloaded_additional_files

    zenodo_metadata, archive_url = zenodo_response
    # Compare the latest downloaded additional files with the latest Zenodo version
    if zenodo_metadata["version"] == latest_downloaded_additional_files.split("_v")[1]:
        logger.info(
            f"Latest additional files ({latest_downloaded_additional_files}) are already present."
        )
        return ADDITIONAL_FILES_FOLDER / latest_downloaded_additional_files

    logger.info(
        f"New additional files available on Zenodo ({zenodo_metadata['version']}). Deleting old and fetching new additional files..."
    )
    # delete old additional files
    shutil.rmtree(
        ADDITIONAL_FILES_FOLDER / latest_downloaded_additional_files,
        onerror=lambda func, path, excinfo: logger.warning(
            f"Failed to delete {path}: {excinfo}"
        ),
    )
    return _download_additional_files(
        zenodo_metadata=zenodo_metadata, record_id=record_id, archive_url=archive_url
    )


def _get_latest_version_folder_name(folders: List[Path]) -> str | None:
    """Get the latest (non empty) version folder name from the list of folders.

    Args:
        folders (List[Path]): List of folders matching the pattern.

    Returns:
        str | None: Latest version folder name if one exists, else None.
    """
    if not folders:
        return None
    latest_downloaded_folder = sorted(
        folders,
        reverse=True,
        key=lambda x: tuple(map(int, str(x).split("_v")[1].split("."))),
    )[0]
    # check folder is not empty
    if not list(latest_downloaded_folder.glob("*")):
        return None
    return latest_downloaded_folder.name


def _get_zenodo_metadata_and_archive_url(record_id: str) -> Tuple[Dict, str] | None:
    """Get the metadata for the Zenodo record and the files archive url.

    Args:
        record_id (str): Zenodo record ID.

    Returns:
        Union[Tuple[Dict,str], None]: Tuple with Metadata dict and archive URL if the record exists, else None.
    """
    try:
        response = requests.get(f"{ZENODO_RECORD_BASE_URL}/{record_id}")
        if response.status_code != 200:
            error_msg = f"Can not find additional files for record_id '{record_id}' on Zenodo ({response.status_code=}). Exiting..."
            logger.error(error_msg)
            raise ZenodoException(error_msg)

        data = response.json()
        return data["metadata"], data["links"]["archive"]

    except requests.exceptions.RequestException as e:
        logger.warning(f"Failed to fetch Zenodo metadata: {e}")
        return None


def _download_additional_files(
    zenodo_metadata: Dict, record_id: str, archive_url: str
) -> Path:
    """Download the latest additional files from Zenodo for the requested record and extract them to the target folder.

    Args:
        zenodo_metadata (Dict): Metadata for the Zenodo record.
        record_id (str): Zenodo record ID.
        archive_url (str): URL to the archive file.

    Raises:
        ZenodoException: If the download fails or the response is not successful.

    Returns:
       Path: Path to the folder where the additional files were extracted.
    """
    record_folder = (
        ADDITIONAL_FILES_FOLDER / f"{record_id}_v{zenodo_metadata['version']}"
    )
    # ensure folder exists
    record_folder.mkdir(parents=True, exist_ok=True)

    logger.info(f"Downloading additional files from Zenodo. This might take a while...")
    # Make a GET request to the URL
    response = requests.get(archive_url, stream=True)
    # Ensure the request was successful
    if response.status_code != 200:
        error_msg = (
            f"Failed to download additional files. Status code: {response.status_code}"
        )
        logger.error(error_msg)
        raise ZenodoException(error_msg)

    _extract_archive(response=response, record_folder=record_folder)

    logger.info(f"Zip file extracted successfully to {record_folder}")
    return record_folder


def _extract_archive(response: requests.Response, record_folder: Path):
    # Download with progress bar
    chunk_size = 1024  # 1KB
    bytes_io = BytesIO()

    with Progress(
        SpinnerColumn(),
        TextColumn("[cyan]Downloading additional files..."),
        TextColumn("[cyan]{task.completed:.2f} MB"),
        transient=True,
    ) as progress:
        task = progress.add_task("", total=None)  # Indeterminate progress

        for data in response.iter_content(chunk_size=chunk_size):
            bytes_io.write(data)
            progress.update(
                task, advance=len(data) / (chunk_size**2)
            )  # Convert bytes to MB

    # Extract the downloaded zip file to the target folder
    with zipfile.ZipFile(bytes_io) as zip_ref:
        zip_ref.extractall(record_folder)

    # check if the extracted file is still a zip
    for f in record_folder.iterdir():
        if f.is_file() and f.suffix == ".zip":
            with zipfile.ZipFile(f) as zip_ref:
                files = zip_ref.namelist()
                with Progress(transient=True) as progress:
                    task = progress.add_task(
                        "[cyan]Extracting files...", total=len(files)
                    )
                    # Iterate over the files and extract them
                    for i, file in enumerate(files):
                        zip_ref.extract(file, record_folder)
                        # Update the progress bar
                        progress.update(task, completed=i + 1)
            f.unlink()  # remove zip after extraction



from loguru import logger

from brats.core.inpainting_algorithms import Inpainter
from brats.core.missing_mri_algorithms import MissingMRI
from brats.core.segmentation_algorithms import (
    AdultGliomaPreTreatmentSegmenter,
    AdultGliomaPreAndPostTreatmentSegmenter,
    AfricaSegmenter,
    GoATSegmenter,
    MeningiomaSegmenter,
    MetastasesSegmenter,
    MeningiomaRTSegmenter,
    PediatricSegmenter,
)

logger.remove()
